
/* kohonen.cpp
	KOHONEN SELF-ORGANIZING NEURAL NETWORK

 Actualizacion 2-2-1993 */

/* #include <conio.h> */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <time.h>
#include "vector.h"
#include "graphnet.h"


/* Prototipos de las funciones utilizadas */

void help(); 
void DisplayNet();
void ResetScreen();
void readweights(char *szWeights);
void readpatterns(char *szPattern);
void writeweights(char *szWeights);
void writeactivations(char *szResults);
int findwinner(int p);
void normalizeweights(int j);

#define min(x,y)	((x)<(y)?(x):(y))
#define max(x,y)	((x)>(y)?(x):(y))
#define TRUE 1
#define FALSE 0
#define ABORT 17 /* ^Q */
#define ESC 27
#define ITEMS 8

/* Definicion de las matrices de almacenamiento dinamico */

MATRIX  out0;                 /* input layer */
MATRIX  out2;                  /* output layer  */
MATRIX  Weights;               /* weights for imput to output layer */
VECTOR  PatternID;            /* identifier for each pattern */
int nInputNodes;
int nOutputNodes;
int nPatterns;
int nIterations;
int p;         /* indice del numero de patron */
int Side;      /* tamano del lado del rectangulo de salida*/
long Tk;       /* periodo de tiempo en pasos de timepo */
float eta;     /* parametro de aprendizaje actual */
int DensityFunction; /* probabilidad de densidad de los patrones */
int q;   /* cuenta de intentos */
int t;   /* periodo de tiempo en segmento actual  */


/* main programc  Kohonen Self_organizing Neural Network */

void main(int argc, char *argv[])
{

float   eta0 = (float) 0.3;           /* default learning rate */
float etaShrink0 =(float) 0.2;
register int i;              /* index hidden layer   */
register int j;             /* index input layer   */
int r;                      /* index output layer */
int n;                     /* numero de capas intermedias */
int nRuns;                /* number of input nodes   */
int T1=1000;
int T1f= 5;
int T2= 100;
char MonitorNet = 0;
int Neighborhood0;
float NeighborSize =(float) 0.5;
int Neighborhood;
FILE *fpRun;              /* run file  */
char  szResults[66];          /*  */
char  szPattern[66];
char  szWeights[66];
char  szWeightsOut[66];
char  *programe = * argv;
int WinnerHorizontal;
int WinnerVertical;
int left;
int right;
int top;
int bottom;
int h;
int v;

/* Read arguments from DOS command line.*/

for (; argc > 1; argc--)
{
	char *arg = *++argv;
	if (*arg != '-')
		 break;
	switch (*++arg)
	{
		case 'e':   sscanf(++arg,  "%d",  &T2);  break;
		case 't':   sscanf(++arg,  "%d",  &T1); break;
		case 'T':   sscanf(++arg,  "%d",  &T1f);  break;
		case 'n':   sscanf(++arg,  "%f",  &NeighborSize); break;
		case 'a':   sscanf(++arg,  "%f",  &etaShrink0);  break;
		default:    break;
	}
}

if (argc < 2)
	help();



/* open run file for reading, */   

if ((fpRun = fopen(*argv, "r")) == NULL)
{
	fprintf(stderr,"%s: Can't open file %s \n",programe,*argv);
	exit(1);
}



fscanf(fpRun, "%d",&nRuns); /* scan for numbers of runs, */ 

	/*******************************************/
	/* Comienzo del bucle de trabajo Principal */
	/*******************************************/

for (r= 0; r< nRuns; r++)
{
/* Read and parse the run specification line to obtain */
/* information about this network. */


	/***************************************************************/
	/*  Los parametros del fichero ejecucion son :		       */
	/*	Fichero de resultados finales			       */
	/*	Fichero de patrones de entreneamiento o testeoPor      */
	/*      Fichero de pesos iniciales (RAMDOM)                    */
	/*	Fichero de pesos de salida    		               */
	/*	Numero de patrones 				       */
	/*      Numero de iteraciones                                  */
	/*	Numero de nodos de entrada			       */
	/* 	Numero de nodos de salida   			       */
	/*	Parametro de aprendizaje (eta)			       */
	/***************************************************************/
	


fscanf(fpRun, "%s %s %s %s %d %d %d %d %f",
	 szResults,                /* output results file */
	 szPattern,                /* pattern input file  */
	 szWeights,                /* initial weights file   */
	 szWeightsOut,             /* final weights output filt  */
	 &nPatterns,               /* number of patterns to lea  */
	 &nIterations,             /* number of iterations the data */
	 &nInputNodes,             /* number of input nodes   */
	 &nOutputNodes,            /* number Of Output nodeS  */
	 &eta);                   /* learning rate          */


   /* La topologia de la capa de salida debe ser rectangular */

	Side = (int) sqrt ( (double) nOutputNodes);
   n = Side * Side;
   if (n!= nOutputNodes )
	{
	fprintf(stderr,"Se usara una capa de salida de %d por %d\n",n,n);
	nOutputNodes = n;
	} /* de if */

/* Allocate dynamic storage for nodes and patterns. */

AllocateMatrix(&out0,  nPatterns,     nInputNodes);
AllocateMatrix(&out2,  nPatterns,     nOutputNodes);
AllocateMatrix(&Weights, nOutputNodes, nInputNodes);
AllocateVector(&PatternID,nPatterns);

 /*  Read the initial Weights matrices   */
readweights(szWeights);

 /* Read in all patterns to be learned, */
readpatterns(szPattern);

fprintf(stderr, nIterations > 1 ? "Training...\n" : "Testing\n");

	/*******************/
	/* ITERRATION LOOP */
	/*******************/
/* Inicializar los valores iniciales */
/* eta = eta0; */
Neighborhood0 =  (int) (Side * NeighborSize);
Tk = 0;
t = 0;                        

	/***************************************************************/
	/*  Partes basicas de las operaciones en el lapso de tiempo t; */
	/*	este bucle interior consta de T1 lapsos de tiempo;     */
	/*	Por cada lapso de tiempo:                              */
	/*		.- un patron se presenta.                      */
	/*		.- un ganador se selecciona.                   */
	/*		.- los pesos de los vecinos son adaptados.     */
	/*                                                             */
	/* Cada valor de q es un lapso (conjunto completo de patrones) */
	/*     							       */	
	/***************************************************************/
	
	
for (q = 0; q < nIterations && eta > 0; q++)         /* numero de iteraciones */
{
  for (p = 0; p < nPatterns; p++)        /* numero de patrones   */
	{
	int Winner;

	/* calcula la activacion de la capa de salida  */

	for (j = 0; j < nOutputNodes; j++)
	{
		float sum = 0;
		for (i = 0; i < nInputNodes; i++)
		sum   +=   Weights[j][i]  *  out0[p][i];

		/* Compute output. */

		out2[p][j] = sum;
	} /* de j */

/* Solo queremos el valor del nodo ganador pero esto lo
	dejamos para luego */

/* buscamos el nodo ganador  */

	Winner = findwinner(p);
	

/* Adaptamos los pesos de imput:output */
/* Definimos los vecinos :
	Solo los pesos de estas unidades vecines de la ganadora seran ajustados */

/* Coordenadas Horizontales y Verticales  del ganador en
	 un espacio bidimensional */

	WinnerHorizontal = Winner % Side;
	WinnerVertical = Winner / Side;

/* La anchura de la vecindad es una funcion linear discreta de lapso de
	tiempo t */

	Neighborhood = Neighborhood0 * ( 1 - t/T1 ) + 1;

/* Restricciones impuestas por array */
	left = max((0),(WinnerHorizontal - Neighborhood));
	right = min((Side -1),(WinnerHorizontal + Neighborhood));
	top = max((0),(WinnerVertical - Neighborhood));
	bottom = min((Side -1),(WinnerVertical + Neighborhood));

/* el coeficiente de apredizaje es una funcion linear discreta de lapso de
	tiempo t */

	eta = eta0 * ( 1 - ((float)t/(float)T1) );
	
/* Modificar los pesos de los vecinos proximos al ganador */

	for (h = left;  h <= right;  h++)
		for (v = top; v <= bottom; v++)
		{
		 j = Side * v + h;
/* Cada componente del vector de pesos cambia alguna fraccion,
	eta de la diferencia con el vector de entradas */

		 for (i = 0; i < nInputNodes; i++)
			Weights[j][i] += eta * (out0[p][i] -Weights[j][i] );
		}
		
		 /* de v */
		
		
		/*************************/
		/* Peticiones de teclado */
		/*************************/
        

		 
	/*if (kbhit())
	  {        
	  
	    int c = toupper(getch());
		if (c=='E')
			MonitorNet++;    Peticion para ver el estado de la red 
		else if (c== ESC)
			goto DONE;   	 Finaliza  grabando 
		else if (c==ABORT)
			goto QUIT;	Finaliza sin almacenar ficheros  
	  } */
	    

	Tk++;
	if (++t >= T1)
	{
			/* Se ha completado un segmento linear de tiempo, T1 */
		t =0;                   /* Comienzo de un nuevo segmento */
		eta0 *= etaShrink0;     /* Reduce el factor de aprendizaje */
		Neighborhood0 = 0;      
		T1 *= T1f;              /* Ajusta el tamao del siguiente segmento */
	} /* de if */
 } /* de p */

 
if (q % T2 == 0 )
   	printf("Interacion %d con un valor de Eta= %f \n", q ,eta);


		/******************************/
		/* Muestra el estdo de la Red */
		/******************************/

 if (MonitorNet || (q % T2 == 0))
	 {
	 	DisplayNet();  
		MonitorNet = 0;
	 }	/* de if */			
} /* de q */

	/***********************/
	/* END ITERRATION LOOP */
	/***********************/

/*DONE:*/

DisplayNet();
 
/*	for(;;);
	if (kbhit()!=0)
	{             
		int c = getch();
		if (c== ESC || c == ABORT)
			break;
	} */ /* de if */

ResetScreen();

/* Escribe los pesos finales */

writeweights(szWeightsOut);

/* Escribe los valores de activacion finales */

writeactivations(szResults);

/* Libera la memoria. */

  FreeMatrix(out0, nPatterns);
  FreeMatrix(out2, nPatterns);
  FreeMatrix(Weights, nOutputNodes);
  free(PatternID);

 } /* de r (numero de ejcuciones) */

/*QUIT:

	fprintf(stderr, " Programa abortado\n");  
	fprintf(stderr, " No se han almacenado los valores en los ficheros\n "); */

	fclose(fpRun);   /* Cerramos el fichero de ejecuciones */
    ResetScreen();

} /* de main */

	/****************************************************/
	/********   FIN DEL PROGRAMA PRINCIPAL  *************/   
	/****************************************************/


	/********************************************************/
	/* AYUDAS PARA EVITAR ERRORES EN EL PASO DE PARAMETROS  */
	/********************************************************/

void help()
{
  fprintf(stderr, "Usage: kohonen {-etTna} runfilename\n");
  fprintf(stderr, " -en => report error every n iterations (100)\n");
  fprintf(stderr, " -tn => initial learning time interval steps (1000)\n");
  fprintf(stderr, " -Tn => learning time interval increase factor (5)\n");
  fprintf(stderr, " -nf => intial neighborhood size fraction of net (0.5)\n");
  fprintf(stderr, " -af => learning rate decreasing factor (.2)\n");
  exit(1);
} /* de help */
  
  /********************************************************/
  /*   BUSCA LA UNIDAD GANADORA PARA UN PATRON DE ENTRADA */
  /********************************************************/

int findwinner(int p)
{
	int Winner;
	int i;
	int j;
	float MinimunDistance = 0;
	for (j = 0 ; j < nOutputNodes; j++)
	{
		float EuclideanDistance = 0;
		for (i = 0 ; i < nInputNodes; i++)
		{
			float d = out0[p][i] - Weights[j][i];
			EuclideanDistance += d * d;
		}
		if (j==0 || EuclideanDistance <= MinimunDistance)
		{
			MinimunDistance = EuclideanDistance;
			Winner = j;
		}
	}
	return Winner;
} /* de findwinner */

  /********************************************************/
  /*    LEE LOS PESOS DE LAS CONNEXIONES                  */
  /********************************************************/

  void readweights(char *szWeights)
  {
	int i;
	int j;
	if (strcmp(szWeights,"RAMDOM") == 0)
	{
		/* pesos aleatorios */
		srand((unsigned) time(NULL));
		for (j = 0 ; j < nOutputNodes; j++)
		{
			for (i = 0 ; i < nInputNodes; i++)
			{
				float r = (float)rand() / RAND_MAX; /* 0.0 .. 1.0 */
				Weights[j][i] = r;
			}
			normalizeweights(j);
		}
	}
	else
	{
		FILE *fpWeights;    /*  initial weight file  */
		if (( fpWeights = fopen(szWeights,"r")) == NULL)
		{
			fprintf(stderr, "can't open file %s\n",szWeights);
			exit(1);
		}
		/*  Leer los pesos de las conneciones imput : output */

		 for (j = 0 ; j < nOutputNodes; j++)
		 {
			for (i = 0 ; i < nInputNodes; i++)
			{
				fscanf(fpWeights, "%f", &Weights[j][i]);
			}
			normalizeweights(j);
		 }
		 fclose(fpWeights);
	}
  }

  /********************************************************/
  /*    NORMALIZA EL VECTOR DE PESOS                      */
  /********************************************************/


  void normalizeweights(int j)
  {
	int i;
	float d = 0;
	for (i = 0 ; i < nInputNodes; i++)
	{
		float r = Weights[j][i];
		d += r * r;
	}
	d = sqrt((double) d);

	for (i = 0 ; i < nInputNodes; i++)
		Weights[j][i] /= d;

  }

  /********************************************************/
  /*                                                      */
  /********************************************************/

  void readpatterns(char *szPattern)
  {
	int i;
	if (strncmp(szPattern,"RAMDOM/",7)==0)
	{
		char c = szPattern[7];
		/* Genera la distribucion de entrada aleatoriamente */
		c = toupper(c);
		if (strchr("TCS",c)==NULL)
		{
			fprintf(stderr,"No entiendo RAMDOM/%c %s \n",c,szPattern);
			exit(1);
		}
		DensityFunction = c;
		for(p = 0; p < nPatterns ; p++)
		{
			float n0, n1;
			for(;;)
			{
				int Inside = FALSE;
				/* genera un conjunto de muestras aleatorio */
				n0 = (float)rand() / RAND_MAX;
				n1 = (float)rand() / RAND_MAX;
				switch(c)
				{
					case 'S':
						/* Distribucion cuadrada */
						Inside = TRUE;
						break;
					case 'T':
						/* Distribucion triangular */
						if (n1 >= 2 * fabs((double)n0 -0.5))
							Inside = TRUE;
						break;
					case 'C':
						/* Distribucion forma cruzada */
						if (fabs((double)n0 -0.5) <= .25 || fabs((double)n1 -0.5) <= .25 )
								Inside = TRUE;
						break;
				}
				if (Inside)
					break;
			}
			out0[p][0] = n0;
			out0[p][1] = n0;
			PatternID[p] = p;
		}
	}
	else
	{
		FILE *fpPattern;    /*  fichero de los patrones de entrada  */
		if (( fpPattern = fopen(szPattern,"r")) == NULL)
		{
			fprintf(stderr, "can't open file %s\n",szPattern);
			exit(1);
		}
		/*  Leer los pesos de las conneciones imput : output */

		 for (p = 0 ; p < nPatterns; p++)
		 {
			float d = 0;
			for (i = 0 ; i < nInputNodes; i++)
			{
				float r;
				if (fscanf(fpPattern, "%f", &r) != 1)
					goto ALLPATTERNSREAD;
				out0[p][i] = r;
				d += r * r;
			}
			d = sqrt((double)d); 
			
		/* Normaliza los Patrones de Entrada */ 
		
			for (i = 0; i < nInputNodes; i++)
				out0[p][i] /=d;
			/*fscanf(fpPattern, "%f", &PatternID[p]);*/
		 }
		 ALLPATTERNSREAD:
		 fclose(fpPattern);

		 if (p < nPatterns)
		 {
			fprintf(stderr,"%d  de  %d patrones leidos \n",p,nPatterns);
			nPatterns = p;
		 }
	}
  }

  /********************************************************/
  /*    FUNCION QUE ESCRIBE LOS PESOS                     */
  /********************************************************/
	void writeweights(char *szWeightsOut)
	{
		int i;
		int j;
		FILE * fpWeightsOut;
		if (( fpWeightsOut = fopen(szWeightsOut,"w")) == NULL)
		{
			fprintf(stderr, "can't open file %s\n",szWeightsOut);
			exit(1);
		}
		for (j = 0 ; j < nOutputNodes; j++)
			for (i = 0 ; i < nInputNodes; i++)
				fprintf(fpWeightsOut, "%f\n",Weights[j][i]);
		fclose(fpWeightsOut);
	}

  /********************************************************/
  /*  ESCRIBE EL VECTOR DE SALIDA FINAL                   */
  /********************************************************/
	void writeactivations(char *szResults)
	{
		int j;
		FILE * fpResults;
		if (( fpResults = fopen(szResults,"w")) == NULL)
		{
			fprintf(stderr, "can't open file %s\n",szResults);
			fpResults = stderr;
		}
		for (p = 0 ; p < nPatterns; p++)
		{
			for (j = 0 ; j < nOutputNodes; j++)
				fprintf(fpResults, "%f\n",out2[p][j]);
/*			fprintf(fpResults, " \n"); */
/*			fprintf(fpResults, "%  6.0f\n",PatternID[p]); */
		}
		fclose(fpResults);
	}


  /********************************************************/
  /*                                                      */
  /********************************************************/
                                                        
                                                        
   
  /********************************************************/
  /*                                                      */
  /********************************************************/
/* Array storage allocation rutines */

  void AllocateVector(VECTOR *Vector, int nCols)
  {
	 if ((*Vector = (VECTOR) calloc(nCols, sizeof(float))) == NULL)
	  {
		 fprintf(stderr, " Sorry Not enough memory for nodes !\n");
		 exit(1);
	  }
  }

  /* Allocate space for dynamic, two-dimensii */
  void AllocateColumns(PFLOAT Matrix[], int nRows, int nCols)
  {
	  int i;
	  for (i = 0; i < nRows; i++)
		 AllocateVector(&Matrix[i], nCols);
  }

void AllocateMatrix(MATRIX *Pmatrix,int nRows, int nCols)
{
	if ((*Pmatrix = (MATRIX) calloc(nRows, sizeof(PFLOAT))) ==
		 NULL)
	{
		fprintf(stderr, " Sorry Not enough memory for nodes !\n");
		exit(1);
	}
	AllocateColumns(*Pmatrix, nRows, nCols);
}
/* Free the memory used by the matrix. */


void FreeMatrix(MATRIX Matrix, int nRows)
{
int i;
for (i = 0; i < nRows; i++)
  free(Matrix[i]);
free(Matrix);
}

void  ResetScreen()
{
}
                      
void DisplayNet()
{
}
